"""
Dependency object to handle multi dependency managers.
"""
import logging

LOGGER = logging.getLogger(__name__)


class Dependency(object):
    """Simple container to keep track of all the required
    info to install a dependency.
    """

    def __init__(self, name, version=None, pkg_mng=None, channel=None, intent='install'):
        self.name = name
        """name of dependency"""

        self.version = version
        """expected version"""

        self.package_manager = pkg_mng
        """either conda, pip or git url"""

        self.channel = channel
        """place to find the dependency depends on package_manager"""

        if isinstance(intent, str):
            intent = {intent}

        self.intents = set(intent)
        """When does the package require this option:
         - 'install': only for installation (prod purpose)
         - 'test': for testing purpose only
         - 'doc': to compile the associated documentation
         - 'example': to replay the examples
         - 'pkglts': used internally to create dependencies between pkglts options
        """

    def __str__(self):
        return "dep: {}".format(self.name)

    def _conda_fmt_name(self):
        if self.version is None:
            version = ""
        else:
            version = self.version
            if version[:2] in ('==', '>=', '<=', "~="):
                LOGGER.warning("bad version specification for '%s' with conda, use '=' by default", self.name)
                version = "=" + version[2:]
            elif version[0] != '=':
                version = "=" + version
        return "{}{}".format(self.name, version)

    def _pip_fmt_name(self):
        if self.version is None:
            version = ""
        else:
            version = self.version
            if version[:2] in ('==', '>=', '<=', "~="):
                # full formatter already do nothing
                pass
            elif version[0] == '=':
                # transform into '=='
                version = "=" + version
            else:
                # no formatter at all use heuristic to include one
                if len(version.split(".")) == 1:
                    major = int(version)
                    next_version = "{:d}".format(major + 1)
                    version = ">=" + version + ", <" + next_version
                elif len(version.split(".")) == 2:
                    major, minor = (int(v) for v in version.split("."))
                    next_version = "{:d}.{:d}".format(major, minor + 1)
                    version = ">=" + version + ", <" + next_version
                else:
                    version = "==" + version
        return "{}{}".format(self.name, version)

    def conda_full_name(self):
        """Produce fully qualified name with version number.

        Returns:
            (str)
        """
        return self._conda_fmt_name()

    def pip_full_name(self):
        """Produce fully qualified name with version number.

        Returns:
            (str)
        """
        if self.is_pip(strict=False):
            return self._pip_fmt_name()

        if self.is_conda(strict=True):
            return self._conda_fmt_name()

        # TODO assume valid git url
        return "walou {}".format(self.name)

    def is_conda(self, strict=True):
        """Check whether this dependency can be managed by conda

        Args:
            strict (bool): whether only conda can manage this dep

        Returns:
            (bool)
        """
        if strict:
            return self.package_manager == 'conda'

        return self.package_manager is None or self.package_manager == 'conda'

    def is_pip(self, strict=True):
        """Check whether this dependency can be managed by pip

        Args:
            strict (bool): whether only pip can manage this dep

        Returns:
            (bool)
        """
        if strict:
            return self.package_manager == 'pip'

        return self.package_manager is None or self.package_manager == 'pip'

    def conda_install(self):
        """Produce command line needed to install this dependency.

        Returns:
            (str)
        """
        full_name = self.conda_full_name()

        if self.channel is None:
            return "conda install {}".format(full_name)

        return "conda install -c {} {}".format(self.channel, full_name)

    def pip_install(self):
        """Produce command line needed to install this dependency.

        Returns:
            (str)
        """
        full_name = self.pip_full_name()
        return "pip install {}".format(full_name)

    def fmt_conda_requirement(self):
        """Format dependency for conda requirements.yml files

        Returns:
            (str)
        """
        return self.conda_full_name()

    def fmt_pip_requirement(self, extended=False):
        """Format dependency for pip requirements.txt files

        Returns:
            (str)
        """
        txt = self._pip_fmt_name()
        if extended:
            if self.is_pip(strict=False):
                txt += "  # {}".format(self.pip_install())
            elif self.is_conda(strict=True):
                txt += "  # {}".format(self.conda_install())
            else:  # assume valid git url
                txt += "  # pip install git+{}".format(self.package_manager)

        return txt



"""
Set of function to store the version of each option each time the package is
regenerated and access them later to prevent reverting changes.
"""
import json
import logging

import semver

from .config import pkg_version_file, pkglts_dir
from .option_tools import available_options

LOGGER = logging.getLogger(__name__)


def load_pkg_version(rep="."):
    """Read version file associated to this package.

    Args:
        rep (Path): directory to search for info

    Returns:
        (dict of str, str): option name, version
    """
    with open(rep / pkglts_dir / pkg_version_file, 'r') as fhr:
        return json.load(fhr)


def write_pkg_version(cfg, rep="."):
    """Store version of tools associated to this package on disk.

    Args:
        cfg (Config): current working config
        rep (Path): directory to search for info

    Returns:
        None
    """
    LOGGER.info("write package version")

    ver = {name: option_current_version(name) for name in cfg.installed_options()}

    with open(rep / pkglts_dir / pkg_version_file, 'w') as fhw:
        json.dump(ver, fhw, sort_keys=True, indent=2)


def option_current_version(name):
    """Get current version of a given option.

    Args:
        name (str): name of option

    Returns:
        (str): X.X.X
    """
    return available_options[name].version()


def outdated_options(cfg, rep="."):
    """Find options which actual version is too old.

    An option is outdated if it's current version is older
    than the version used to regenerate the package.

    Args:
        cfg (Config): current working config
        rep (Path): directory to search for info

    Returns:
        (list of str): list of option names
    """
    try:
        ver = load_pkg_version(rep)
    except FileNotFoundError:
        return []

    outdated = []
    for name in cfg.installed_options():
        ver_cur = option_current_version(name)
        ver_last_rg = ver.get(name, "0.0.0")
        if semver.compare(ver_cur, ver_last_rg) < 0:
            outdated.append(name)

    return outdated



from pathlib import Path

from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionReadthedocs(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            project="{{ github.project }}"
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        project = cfg[self._name]['project']

        if not project:
            invalids.append("readthedocs.project")

        return invalids

    def require_option(self):
        return ['pysetup', 'github', 'sphinx']

    def environment_extensions(self, cfg):
        project = cfg['readthedocs']['project']
        project = project.replace(".", "")
        url = f"{project}.readthedocs.io/en/latest/?badge=latest"
        img = f"readthedocs.org/projects/{project}/badge/?version=latest"
        badge = fmt_badge(img, url, "Documentation status", cfg['doc']['fmt'])

        return {"badge": badge}



import json
from pathlib import Path

import pytest
from pkglts.config import pkg_cfg_file, pkglts_dir
from pkglts.config_management import (CURRENT_PKG_CFG_VERSION, Config, DEFAULT_CFG, get_pkg_config, write_pkg_config)
from pkglts.small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_dir():
    pth = Path("toto_mg_cfg")
    ensure_created(pth)
    ensure_created(pth / pkglts_dir)

    yield pth

    if pth.exists():
        rmdir(pth)


def test_create_env():
    cfg = Config(DEFAULT_CFG)
    assert len(tuple(cfg.installed_options())) == 0

    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(a=1, b=2)
    cfg = Config(pkg_cfg)
    assert len(tuple(cfg.installed_options())) == 1
    assert cfg['base']['a'] == 1
    assert cfg['base']['b'] == 2


def test_create_env_render_templates():
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(a="a", b="b")
    pkg_cfg['tpl'] = dict(tpl1="{{ base.a }}",
                          tpl2="{{ base.b }} and {{ tpl.tpl1 }}")

    cfg = Config(pkg_cfg)
    assert cfg['tpl']['tpl1'] == "a"
    assert cfg['tpl']['tpl2'] == "b and a"


def test_create_env_raise_error_if_unable_to_fully_render_templates():
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(a="{{ base.b }}", b="{{ base.a }}")

    with pytest.raises(UserWarning):
        Config(pkg_cfg)


def test_pkg_env_raise_error_if_option_not_defined():
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['babou'] = dict(a="a", b="b")

    with pytest.raises(KeyError):
        Config(pkg_cfg).load_extra()


def test_pkg_env_loads_specific_handlers_from_options():
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname="toto", namespace="nm",
                           url=None, authors=[("moi", "moi@aussi")])

    cfg = Config(pkg_cfg)
    cfg.load_extra()
    assert hasattr(cfg._env.globals['base'], "pkg_full_name")


def test_get_pkg_config_read_cfg(tmp_dir):
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname="toto", namespace="nm",
                           url=None, authors=[("moi", "moi@aussi")])
    json.dump(pkg_cfg, open(tmp_dir / pkglts_dir / pkg_cfg_file, 'w'))

    cfg = get_pkg_config(tmp_dir)
    assert 'base' in cfg


def test_get_pkg_config_handle_versions(tmp_dir):
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg["_pkglts"]["version"] = 0
    json.dump(pkg_cfg, open(tmp_dir / pkglts_dir / pkg_cfg_file, 'w'))

    cfg = get_pkg_config(tmp_dir)
    assert cfg["_pkglts"]['version'] == CURRENT_PKG_CFG_VERSION


def test_pkg_cfg_read_write_maintains_templates(tmp_dir):
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname="toto", namespace="nm",
                           url=None, authors=[("moi", "moi@aussi")])
    pkg_cfg['license'] = dict(name="CeCILL-C", organization="org",
                              project="{{ base.pkgname }}", year="2015")

    json.dump(pkg_cfg, open(tmp_dir / pkglts_dir / pkg_cfg_file, 'w'))

    cfg = get_pkg_config(tmp_dir)
    assert cfg['license']['project'] == "toto"

    write_pkg_config(cfg, tmp_dir)
    pkg_cfg = json.load(open(tmp_dir / pkglts_dir / pkg_cfg_file))
    pkg_cfg["base"]["pkgname"] = "tutu"

    json.dump(pkg_cfg, open(tmp_dir / pkglts_dir / pkg_cfg_file, 'w'))

    cfg = get_pkg_config(tmp_dir)
    assert cfg['license']['project'] == "tutu"



import pytest
from pkglts.config_management import Config
from pkglts.option.landscape.option import OptionLandscape


@pytest.fixture()
def opt():
    return OptionLandscape('landscape')


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 2


def test_require(opt):
    cfg = Config()
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



import json
from os.path import exists
from os.path import join as pj
from pathlib import Path

import pytest
import semver
from pkglts.config import pkg_version_file, pkglts_dir
from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.small_tools import ensure_created, rmdir
from pkglts.version import __version__
from pkglts.version_management import load_pkg_version, option_current_version, outdated_options, write_pkg_version


@pytest.fixture()
def tmp_dir():
    pth = Path("toto_mg_ver")
    ensure_created(pth)
    ensure_created(pth / pkglts_dir)

    yield pth

    rmdir(pth)


@pytest.fixture()
def tmp_cfg():
    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname="toto", namespace="nm",
                           url=None, authors=[("moi", "moi@aussi")])
    pkg_cfg['license'] = dict(name="CeCILL-C", organization="org",
                              project="{{ base.pkgname }}", year="2015")

    cfg = Config(pkg_cfg)
    return cfg


def test_pkglts_options_use_pkglts_version():
    assert option_current_version('base') == __version__


def test_freshly_regenerated_package_have_not_outdated_options(tmp_cfg, tmp_dir):
    write_pkg_version(tmp_cfg, tmp_dir)

    assert len(outdated_options(tmp_cfg, tmp_dir)) == 0


def test_outdated_options_handles_no_version_file(tmp_cfg, tmp_dir):
    assert len(outdated_options(tmp_cfg, tmp_dir)) == 0


def test_old_regenerated_package_have_outdated_options(tmp_cfg, tmp_dir):
    write_pkg_version(tmp_cfg, tmp_dir)
    ver = load_pkg_version(tmp_dir)
    ver['base'] = semver.bump_major(ver['base'])
    json.dump(ver, open(tmp_dir / pkglts_dir / pkg_version_file, 'w'))

    assert 'base' in outdated_options(tmp_cfg, tmp_dir)


def test_old_regenerated_handles_unknown_options(tmp_cfg, tmp_dir):
    write_pkg_version(tmp_cfg, tmp_dir)
    ver = load_pkg_version(tmp_dir)
    del ver['base']
    json.dump(ver, open(tmp_dir / pkglts_dir / pkg_version_file, 'w'))

    assert 'base' not in outdated_options(tmp_cfg, tmp_dir)



"""
Script which can be called before generating the doc
to generate description of files modified by each option.
"""
from pkglts.config_management import Config


def _nn(cfg, pth):
    """Helper function"""
    print("pth:", pth)
    tgt_name = cfg.render(pth)
    if tgt_name.endswith(".tpl"):
        tgt_name = tgt_name[:-4]

    return tgt_name


def _tree(dname, padding, txt):
    """Generate tree ascii representation of a dir"""
    pkg_cfg = dict(base={"namespace": None,
                         "owner": "owner",
                         "pkgname": "pkgname",
                         "url": None},
                   data={"use_ext_dir": False},
                   doc={"fmt": "rst"},
                   plugin_project={"plugin_name": "plugin"},
                   test={'suite_name': "pytest"})

    cfg = Config(pkg_cfg)

    files = sorted(dname.iterdir())

    count = 0
    for pth in files:
        fmt_name = _nn(cfg, pth.name)
        if fmt_name.split(".")[0] != '_':
            count += 1
            txt += padding + '|\n'
            txt += padding + '+-' + fmt_name
            if pth.is_dir():
                txt += "/\n"
                if count == len(files):
                    txt = _tree(pth, padding + ' ' + ' ' * int(len(fmt_name) / 2), txt)
                else:
                    txt = _tree(pth, padding + '|' + ' ' * int(len(fmt_name) / 2), txt)
            else:
                txt += '\n'

    return txt


def fmt_tree(dname):
    """Generate tree ascii representation of a dir

    Args:
        dname (Path): path to directory

    Returns:
        (str)
    """
    return _tree(dname, '', ".\n")



from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.option_tools import available_options
from pkglts.version import __version__

from . import history


class OptionDoc(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            description="belle petite description",
            fmt="rst",
            keywords=[]
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        description = cfg[self._name]['description']
        fmt = cfg[self._name]['fmt']
        # keywords = env.globals['doc'].keywords

        if not description:
            invalids.append("doc.description")

        if fmt not in ('rst', 'md'):
            invalids.append("doc.fmt")

        return invalids

    def require_option(self):
        return ['base']

    def require(self, cfg):
        if cfg[self._name]['fmt'] == 'md':
            yield Dependency('mkdocs', intent='doc')

    def environment_extensions(self, cfg):
        badges = []
        for name in sorted(set(cfg.installed_options()) - {self._name}):
            opt = available_options[name]
            ext = opt.environment_extensions(cfg)
            if 'badge' in ext:
                badges.append(ext['badge'])
            if 'badges' in ext:
                badges.extend(ext['badges'])

        return {"badges": badges}

    def tools(self, cfg):
        del cfg
        yield history.parser_history



""" Some helpers for options
"""
import logging

import pkg_resources

LOGGER = logging.getLogger(__name__)


class EpDict(dict):
    """A simple dictionary to load EntryPoint at access time only"""

    def __getitem__(self, name):
        val = dict.__getitem__(self, name)
        if isinstance(val, pkg_resources.EntryPoint):
            val = val.load()(name)
            self[name] = val

        return val

    def items(self):
        raise NotImplementedError("TODO is needed")

    def values(self):
        raise NotImplementedError("TODO is needed")


available_options = EpDict()
"""Dictionary of currently available option in pkglts. Discovered at run time"""


def find_available_options():
    """Discover all available options.

    Returns:
        (dict of str: Option)
    """
    for ept in pkg_resources.iter_entry_points(group='pkglts'):
        option_name = ept.name
        if option_name not in available_options:
            available_options[option_name] = ept


def get_user_permission(action_name, default_true=True):
    """Helper function to input a yes or no question.

    Args:
        action_name (str): action to ask permission for.
        default_true (bool): whether answer is yes or no by default.

    Returns:
        (bool)
    """
    if default_true:
        ans = input(f"{action_name} [y], n?") in ("", "y")
    else:
        ans = input(f"{action_name} y, [n]?") == "y"

    return ans
